/*	===========================================================================
	AADebug Ver. 0.11
	Copyright (C) 2007 Fourteenforty Research Institute, Inc.
	Coded by Yuji Ukai <yukai@fourteenforty.jp>
	===========================================================================
*/

#include <ntddk.h>

#define		USDRIVERNAME	L"\\Device\\AADEBUG"
#define		USDOSDRIVERNAME	L"\\DosDevices\\AADEBUG"

#define		IOCTL_ID_CUSTOM_ADD			0x00001000		// Add process command from userland by DeviceIoControl
#define		IOCTL_ID_CUSTOM_DELETE		0x00001001		// Delete process command from userland by DeviceIoControl
#define		IOCTL_ID_CUSTOM_HOOKENABLE	0x00001002		// Enable hook
#define		STATUS_ERR_PARAM			0xfffffffe		// Error code - parameter error
#define		STATUS_ERR_DUPLICATED		0xfffffffd		// Error code - specified process is already in list
#define		STATUS_ERR_LISTFULL			0xfffffffc		// Error code - list full
#define		STATUS_ERR_NOTFOUND			0xfffffffb		// Error code - specified process is not found

// Pattern-1 : Process is attached with debugger
static UCHAR PATTERN1[18]={
	0x39,0xBE,0xBC,0x00,0x00,0x00,	// cmp     [esi+0BCh], ebx
	0x0F,0x95,0xC0,					// setnz   al
	0x8B,0x8E,0xB0,0x01,0x00,0x00,	// mov     ecx, [esi+1B0h]
	0x88,0x41,0x02					// mov     [ecx+2], al
};

// Pattern-2 : Process is started with debugger
static UCHAR PATTERN2[15]={
	0x39,0x9E,0xBC,0x00,0x00,0x00,	// cmp     [esi+0BCh], ebx
	0x0F,0x95,0xC0,					// setnz   al
	0x8B,0x4D,0xE4,					// mov     ecx, [ebp-1Ch]
	0x88,0x41,0x02					// mov     [ecx+2], al
};

UCHAR		*p_SetDebugFlag1=NULL;	// Address of pattern-1 code
UCHAR		*p_SetDebugFlag2=NULL;	// Address of pattern-2 code
UCHAR		*pBaseNtoskrnl=NULL;	// Base address of ntoskrnl.exe

#define		MAXNUM_PROC		16		// Maximum number of process list
#define		MAXLEN_PROCNAME	32		// Maximum length of process name

char		target_proclist[MAXNUM_PROC][MAXLEN_PROCNAME];	// Target process list

NTSTATUS	DriverEntry(PDRIVER_OBJECT, PUNICODE_STRING);
NTSTATUS	AADEBUGUnSupportedFunction(IN PDEVICE_OBJECT pDeviceObject,IN PIRP pIrp);
NTSTATUS	AADEBUGDeviceOpen(IN PDEVICE_OBJECT pDeviceObject,IN PIRP pIrp);
NTSTATUS	AADEBUGDeviceClose(IN PDEVICE_OBJECT pDeviceObject,IN PIRP pIrp);
NTSTATUS	AADEBUGDeviceIoControl(IN PDEVICE_OBJECT pDeviceObject,IN PIRP pIrp);
VOID		AADEBUGDriverUnload(IN PDRIVER_OBJECT pDriverObject);

/*
	Enable/Disable BeingDebugged flag
	Called from HookDebugRoutine()
*/
unsigned int __stdcall DebugProcessCheck(unsigned int peb,unsigned int DebugFlag,unsigned int location,unsigned int eprocess)
{
	unsigned char	*p,*p_peb;
	unsigned int	d,i;
	char			procname[MAXLEN_PROCNAME];


	
	DbgPrint("DebugProcessCheck addr=%x peb=%x DebugFlag=%d Location=%d\n",DebugProcessCheck,peb,DebugFlag,location);
	
	
	// Check PEB pointer
	p_peb=(unsigned char *)peb;
	if (MmIsAddressValid(p_peb)==FALSE){
		DbgPrint("DebugProcessCheck Error : PEB(%x) is not accessible.\n",p_peb);
		return 0;
	}
	if (MmIsAddressValid(p_peb+2)==FALSE){
		DbgPrint("DebugProcessCheck Error : PEB(%x/debugging) is not accessible.\n",p_peb+2);
		return 0;
	}

	// If process should not be debugged, clear debug flag in PEB and return
	if (DebugFlag==0){
		*(p_peb+2)=0;
		return 0;
	}
	*(p_peb+2)=1;

	// PsGetProcessImageFileName(p);
	p=(unsigned char *)(eprocess+0x174);
	if (MmIsAddressValid(p)==FALSE){
		DbgPrint("DebugProcessCheck Error : process name is not accessible.\n");
		return 0;
	}

	RtlFillMemory(procname,MAXLEN_PROCNAME,0x00);
	strncpy(procname,p,MAXLEN_PROCNAME-1);
	_strupr(procname);
	DbgPrint("DebugProcessCheck ProcessName=[%s]\n",procname);

	// Check process name
	for (i=0;i<MAXNUM_PROC;i++){

		if (strlen(target_proclist[i])==0) continue;
		DbgPrint("Checking %d:[%s]\n",i,target_proclist[i]);
		
		if (strcmp(target_proclist[i],"*")==0){
			DbgPrint("Target process is found (wildcard)\n");
			*(p_peb+2)=0;
			break;
		}

		if (strcmp(target_proclist[i],procname)==0){
			DbgPrint("Target process is found\n");
			*(p_peb+2)=0;
			break;
		}
	}
	if (i==MAXNUM_PROC)
		DbgPrint("Target process is not found\n");

	return 0;
}

/*
	Hook routine
	This function is called when process is attached with debugger or process is created
*/
void __declspec(naked) HookDebugRoutine()
{
	_asm {
			push eax
			push ebx
			push edx
			push edi
			push esi

			// Check code location of setting debug flag
			mov		ecx,dword ptr [p_SetDebugFlag1]
			mov		eax,[esp+20]	; Return address
			add		ecx,5			; Bytes of call instruction				
			cmp		ecx,eax
			jnz		Location2
			
			// Location-1 (Process is attached with debugger)
			// PAGE:0058846F                 cmp     [esi+0BCh], edi
			// PAGE:00588475                 setnz   al
			// PAGE:00588478                 mov     ecx, [esi+1B0h]
			// PAGE:0058847E                 mov     [ecx+2], al

			mov		edx,1					; Location =1
			xor		eax,eax
			cmp     [esi+0BCh], edi
			setnz   al						; EAX=Debug Flag
			mov     ecx, [esi+1B0h]

			jmp		CheckDebugger

Location2:
			// Location-2 (Process is started with Debuger)
			// PAGE:004B53B7                 cmp     [esi+0BCh], ebx
			// PAGE:004B405C                 setnz   al
			// PAGE:004B405F                 mov     ecx, [ebp+var_1C]
			// PAGE:004B4062                 mov     [ecx+2], al

			mov		edx,2					; Location =2
			xor		eax,eax
			cmp     [esi+0BCh], ebx
			setnz   al						; EAX=Debug Flag
			mov     ecx, [ebp-1Ch]

CheckDebugger:



			push	ecx


			push	esi		//	Pointer to EPROCESS structure 
			push	edx		//	Location (1=Attached, 2=Started)
			push	eax		//  Debug Flag (1=Debugger enabled)
			push	ecx		//	PEB of debugging process
			call	DebugProcessCheck

			pop		ecx

			mov		ebx,dword ptr [p_SetDebugFlag1]
			mov		eax,[esp+20]	; Return address
			add		ebx,5			; Bytes of call instruction
			cmp		ebx,eax
			jnz		Exit_Location2

; Exit_Location1:
			pop esi
			pop edi
			pop edx
			pop ebx
			pop eax
				
			pop	eax
			add	eax,0dh
			jmp	eax
				
Exit_Location2:
	
			pop esi
			pop edi
			pop edx
			pop ebx
			pop eax
				
			pop	eax
			add	eax,10
			jmp	eax					
	};
}

void Exchange64(UCHAR *src,UCHAR *dst)
{
	DbgPrint("src,addr=%x : %x %x %x %x %x %x %x %x\n",src,
		src[0],
		src[1],
		src[2],
		src[3],
		src[4],
		src[5],
		src[6],
		src[7]);

	__asm{
		push	edi
		push	esi
		push	eax
		push	ebx
		push	ecx
		push	edx
			
		mov			esi,dword ptr [src]
		mov			edi,dword ptr [dst]
		mov			eax,[esi]
		mov			edx,[esi+4]

		xor			ebx,ebx
		mov			bl,[edi+3]
		shl			ebx,8
		mov			bl,[edi+2]
		shl			ebx,8
		mov			bl,[edi+1]
		shl			ebx,8
		mov			bl,[edi]

		add			edi,4

		xor			ecx,ecx
		mov			cl,[edi+3]
		shl			ecx,8
		mov			cl,[edi+2]
		shl			ecx,8
		mov			cl,[edi+1]
		shl			ecx,8
		mov			cl,[edi]
		
		lock CMPXCHG8B	[esi]
			
		pop		edx
		pop		ecx
		pop		ebx
		pop		eax
		pop		esi
		pop		edi			
	}
	DbgPrint("src,addr=%x : %x %x %x %x %x %x %x %x\n",src,
		src[0],
		src[1],
		src[2],
		src[3],
		src[4],
		src[5],
		src[6],
		src[7]);	
}					
/*
	Search code chunk for enabling BeingDebug flag
*/
BOOLEAN SearchSetDebugFlag(UCHAR *pBaseNtoskrnl,UCHAR **p1,UCHAR **p2)
{
	ULONG				i,j;
	BOOLEAN				ret_v=TRUE;
	UCHAR				*p;
	UCHAR				*ppSetDebugFlag1=NULL;
	UCHAR				*ppSetDebugFlag2=NULL;

	for (i=0;;i++){

		p=pBaseNtoskrnl+i;
		if (p>=(UCHAR *)0xffffffff || MmIsAddressValid(p)==FALSE || MmIsAddressValid(p+12)==FALSE){
			ret_v=FALSE;
			DbgPrint("AADEBUG DriverEntry Not found. Search end addr=%x pSetDebugFlag1=%x pSetDebugFlag2=%x\n",p,ppSetDebugFlag1,ppSetDebugFlag2);
			break;
		}
			
		if (RtlCompareMemory(p,PATTERN1,sizeof(PATTERN1))==sizeof(PATTERN1)){
			if (ppSetDebugFlag1==NULL){
				ppSetDebugFlag1=p;
				DbgPrint("AADEBUG DriverEntry Found ppSetDebugFlag1=%x\n",ppSetDebugFlag1);
			}else{
				ppSetDebugFlag1=(UCHAR *)0xffffffff;
				DbgPrint("AADEBUG DriverEntry Fatal error- ppSetDebugFlag1 is duplicated\n");
				break;
			}
		}

		if (RtlCompareMemory(p,PATTERN2,sizeof(PATTERN2))==sizeof(PATTERN2)){
			if (ppSetDebugFlag2==NULL){
				ppSetDebugFlag2=p;
				DbgPrint("AADEBUG DriverEntry Found ppSetDebugFlag2=%x\n",ppSetDebugFlag2);
			}else{
				ppSetDebugFlag2=(UCHAR *)0xffffffff;
				DbgPrint("AADEBUG DriverEntry Fatal error- ppSetDebugFlag2 is duplicated\n");
				break;
			}
		}

		if (ppSetDebugFlag1 && ppSetDebugFlag2) break;
	}
	if (ppSetDebugFlag1 == (UCHAR *)0xffffffff
		|| ppSetDebugFlag2 == (UCHAR *)0xffffffff){

			DbgPrint("AADEBUG DriverEntry : Not found. Search end addr=%x\n",p);
			
			ppSetDebugFlag1=0;
			ppSetDebugFlag2=0;
			ret_v=FALSE;
	}

	*p1=ppSetDebugFlag1;
	*p2=ppSetDebugFlag2;

	return ret_v;
}

/*
	DriverEntry
*/
NTSTATUS DriverEntry(IN PDRIVER_OBJECT pDriverObject,IN PUNICODE_STRING pRegistryPath)
{

	NTSTATUS			Status = STATUS_SUCCESS;
	PDEVICE_OBJECT		pDeviceObject = NULL;
	ULONG				i;
	UNICODE_STRING		usDriverName, usDosDeviceName;


 	DbgPrint("AADEBUG DriverEntry v0.11");

	p_SetDebugFlag1=NULL;
	p_SetDebugFlag2=NULL;
	pBaseNtoskrnl=NULL;
	RtlFillMemory(target_proclist,sizeof(target_proclist),0x00);
 	
    RtlInitUnicodeString(&usDriverName, USDRIVERNAME);
    RtlInitUnicodeString(&usDosDeviceName, USDOSDRIVERNAME); 
	IoCreateSymbolicLink(&usDosDeviceName, &usDriverName);

    for(i = 0; i < IRP_MJ_MAXIMUM_FUNCTION; i++)
         pDriverObject->MajorFunction[i] = AADEBUGUnSupportedFunction;
 
	pDriverObject->MajorFunction[IRP_MJ_CREATE] = AADEBUGDeviceOpen;
	pDriverObject->MajorFunction[IRP_MJ_CLOSE]  = AADEBUGDeviceClose;
	pDriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]  = AADEBUGDeviceIoControl;
	pDriverObject->DriverUnload = AADEBUGDriverUnload;

	Status = IoCreateDevice(pDriverObject,0,&usDriverName,FILE_DEVICE_UNKNOWN,0,FALSE,&pDeviceObject);
	
	if(!NT_SUCCESS( Status )){
		DbgPrint("AADEBUG DriverEntry IoCreateDevice() failed.");
		return STATUS_UNSUCCESSFUL;
	}
	
    pDeviceObject->Flags |= 0;
    pDeviceObject->Flags &= (~DO_DEVICE_INITIALIZING);
    
	return STATUS_SUCCESS;
}

NTSTATUS AADEBUGUnSupportedFunction(IN PDEVICE_OBJECT pDeviceObject,IN PIRP pIrp)
{
	DbgPrint("AADEBUGUnSupportedFunction: Entry...\n");

	pIrp->IoStatus.Information = 0;
	pIrp->IoStatus.Status = STATUS_SUCCESS;
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );

	return( STATUS_SUCCESS );
}

NTSTATUS AADEBUGDeviceOpen(IN PDEVICE_OBJECT pDeviceObject,IN PIRP pIrp)
{
	DbgPrint("AADEBUGDeviceOpen: Entry...\n");

	pIrp->IoStatus.Information = 0;
	pIrp->IoStatus.Status = STATUS_SUCCESS;
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );

	return( STATUS_SUCCESS );
}
NTSTATUS AADEBUGDeviceClose(IN PDEVICE_OBJECT pDeviceObject,IN PIRP pIrp)
{
	DbgPrint("AADEBUGDeviceClose: Entry...\n");

	pIrp->IoStatus.Information = 0;
	pIrp->IoStatus.Status = STATUS_SUCCESS;
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );

	return( STATUS_SUCCESS );
}

/*
	DeviceIoControl routine
*/

NTSTATUS AADEBUGDeviceIoControl(IN PDEVICE_OBJECT pDeviceObject,IN PIRP pIrp)
{
	PIO_STACK_LOCATION	pIrpSp;
	ULONG				nFunctionCode;
	UCHAR				*pInputBuffer;
	UCHAR				*pOutputBuffer;
	unsigned int		dwInSize;
	unsigned int		dwOutSize;
	unsigned int		nStatusToReturn;
	unsigned int		dwDataWritten=0;
	char				buf[MAXLEN_PROCNAME];
	ULONG				i;

	DbgPrint("AADEBUGDeviceIoControl: Entry...\n");

	pIrpSp = IoGetCurrentIrpStackLocation(pIrp);
	if (pIrpSp==NULL){
		DbgPrint("AADEBUGDeviceIoControl: IoGetCurrentIrpStackLocation() error.\n");
		return STATUS_UNSUCCESSFUL;
	}
	nFunctionCode=pIrpSp->Parameters.DeviceIoControl.IoControlCode;

    pInputBuffer  = (PCHAR)pIrp->AssociatedIrp.SystemBuffer;
    pOutputBuffer = (PCHAR)pIrp->AssociatedIrp.SystemBuffer;
    dwInSize  = pIrpSp->Parameters.DeviceIoControl.InputBufferLength;
    dwOutSize = pIrpSp->Parameters.DeviceIoControl.OutputBufferLength;


	switch(nFunctionCode){
			
		case IOCTL_ID_CUSTOM_HOOKENABLE:		
			{
				UCHAR				*p;
				UCHAR				ib[8];

				// Is specified buffer accessible ?
				if (MmIsAddressValid(pInputBuffer)==FALSE){
					DbgPrint("AADEBUGDeviceIoControl: IOCTL_ID_CUSTOM_ADD : Parameter (%x) is not accessible.\n",pInputBuffer);
					nStatusToReturn = STATUS_ERR_PARAM;
					break;
				}
				pBaseNtoskrnl=(UCHAR *)*((ULONG *)pInputBuffer);
				DbgPrint("AADEBUGDeviceIoControl : IOCTL_ID_CUSTOM_HOOKENABLE : ntoskrnl.exe base=%x\n",pBaseNtoskrnl);
				
				
				// Get base pointer of ntoskrnl.exe
				DbgPrint("AADEBUGDeviceIoControl : IOCTL_ID_CUSTOM_HOOKENABLE : Searching instructions...\n");

				if (SearchSetDebugFlag(pBaseNtoskrnl,&p_SetDebugFlag1,&p_SetDebugFlag2)==TRUE){

					DbgPrint("AADEBUGDeviceIoControl : IOCTL_ID_CUSTOM_HOOKENABLE : Hook enabling...\n");

					// hook

					//------------- Location-1 (process is attached with debugger) ---------------------------
					
					p=p_SetDebugFlag1;
					ib[0]=0xe8;
					*((unsigned int *)(ib+1))=(unsigned int)HookDebugRoutine-(unsigned int)p-5;
					memset(ib+5,0x90,3);
					Exchange64(p_SetDebugFlag1,ib);
						
					//------------- Location-2 (process is started with debugger) ---------------------------
					
					p=p_SetDebugFlag2;
					ib[0]=0xe8;
					*((unsigned int *)(ib+1))=(unsigned int)HookDebugRoutine-(unsigned int)p-5;
					memset(ib+5,0x90,3);
					Exchange64(p_SetDebugFlag2,ib);	
				}

				DbgPrint("AADEBUGDeviceIoControl : IOCTL_ID_CUSTOM_HOOKENABLE : pBaseNtoskrnl=%x p_SetDebugFlag1=%x p_SetDebugFlag2=%x\n",
					pBaseNtoskrnl,p_SetDebugFlag1,p_SetDebugFlag2);
			}
			break;
	
		case IOCTL_ID_CUSTOM_ADD:
			{
				DbgPrint("AADEBUGDeviceIoControl: IOCTL_ID_CUSTOM_ADD dwInSize=%u dwOutSize=%u\n",dwInSize,dwOutSize);

				// Is specified buffer accessible ?
				if (MmIsAddressValid(pInputBuffer)==FALSE){
					DbgPrint("AADEBUGDeviceIoControl: IOCTL_ID_CUSTOM_ADD : Parameter (%x) is not accessible.\n",pInputBuffer);
					nStatusToReturn = STATUS_ERR_PARAM;
					break;
				}

				RtlFillMemory(buf,MAXLEN_PROCNAME,0x00);
				strncpy(buf,pInputBuffer,MAXLEN_PROCNAME-1);
				_strupr(buf);

				// Is specified process already in the list ?
				for (i=0;i<MAXNUM_PROC;i++)
					if (strcmp(target_proclist[i],buf)==0){
						DbgPrint("AADEBUGDeviceIoControl: IOCTL_ID_CUSTOM_ADD : Specified process (%s) is already in list\n",buf);
						break;
					}
				if (i!=MAXNUM_PROC){
					nStatusToReturn = STATUS_ERR_DUPLICATED;
					break;
				}

				// Is proc list full ?
				for (i=0;i<MAXNUM_PROC;i++)
					if (strlen(target_proclist[i])==0)
						break;
				if (i==MAXNUM_PROC){
					DbgPrint("AADEBUGDeviceIoControl: IOCTL_ID_CUSTOM_ADD : List is full\n");
					nStatusToReturn = STATUS_ERR_LISTFULL;
					break;
				}

				// Set entry
				strncpy(target_proclist[i],buf,MAXLEN_PROCNAME-1);
				DbgPrint("AADEBUGDeviceIoControl: Process [%d:%s] is added.\n",i,buf);
				nStatusToReturn = STATUS_SUCCESS;
			}
			break;
			
		case IOCTL_ID_CUSTOM_DELETE:
	   	  	{
				// Is specified buffer accessible ?
				if (MmIsAddressValid(pInputBuffer)==FALSE){
					DbgPrint("AADEBUGDeviceIoControl: IOCTL_ID_CUSTOM_ADD : Parameter (%x) is not accessible.\n",pInputBuffer);
					nStatusToReturn = STATUS_ERR_PARAM;
					break;
				}

				RtlFillMemory(buf,MAXLEN_PROCNAME,0x00);
				strncpy(buf,pInputBuffer,MAXLEN_PROCNAME-1);
				_strupr(buf);

				// Is specified process in the list ?
				for (i=0;i<MAXNUM_PROC;i++)
					if (strcmp(target_proclist[i],buf)==0){
						strcpy(target_proclist[i],"");
						break;
					}
				if (i==MAXNUM_PROC){
					DbgPrint("AADEBUGDeviceIoControl: IOCTL_ID_CUSTOM_ADD : Specified process (%s) is not found in list\n",buf);
					nStatusToReturn = STATUS_ERR_NOTFOUND;
					break;
				}

				DbgPrint("AADEBUGDeviceIoControl: Process [%d:%s] is deleted.\n",i,buf);
				nStatusToReturn = STATUS_SUCCESS;
			}
			break;
			
		default:
			DbgPrint("AADEBUGDeviceIoControl: Invalid function number 0x%x\n",nFunctionCode);
			nStatusToReturn=STATUS_UNSUCCESSFUL;
			break;
	}

	pIrp->IoStatus.Information = dwDataWritten;
	pIrp->IoStatus.Status = nStatusToReturn;
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );

	return( nStatusToReturn );
}

/*
	Unload driver
*/
VOID AADEBUGDriverUnload(IN PDRIVER_OBJECT pDriverObject)
{
	UNICODE_STRING usDosDeviceName;

	DbgPrint("AADEBUGDriverUnload: Entry...\n");



	if (p_SetDebugFlag1){
		Exchange64(p_SetDebugFlag1,PATTERN1);
	}
	if (p_SetDebugFlag2){
		Exchange64(p_SetDebugFlag2,PATTERN2);
	}


	DbgPrint("AADEBUGDriverUnload: Completed...\n");

	RtlInitUnicodeString(&usDosDeviceName, USDOSDRIVERNAME);
	IoDeleteSymbolicLink(&usDosDeviceName);
	IoDeleteDevice(pDriverObject->DeviceObject);
}
