/*	===========================================================================
	AADebug Ver. 0.11
	Copyright (C) 2007 Fourteenforty Research Institute, Inc.
	Coded by Yuji Ukai <yukai@fourteenforty.jp>
	===========================================================================
*/
#include <windows.h>
#include <stdio.h>
#include "drvload.h"

#define	DRIVERNAME		"AADEBUG.sys"
#define	DEVICENAME		"AADEBUG"

#define		IOCTL_ID_CUSTOM_ADD			0x00001000
#define		IOCTL_ID_CUSTOM_DELETE		0x00001001
#define		IOCTL_ID_CUSTOM_HOOKENABLE	0x00001002
#define		STATUS_ERR_PARAM			0xfffffffe
#define		STATUS_ERR_DUPLICATED		0xfffffffd
#define		STATUS_ERR_LISTFULL			0xfffffffc
#define		STATUS_ERR_NOTFOUND			0xfffffffb
#define		MAXNUM_PROC		16
#define		MAXLEN_PROCNAME	32

#define		SystemModuleInformation 11

void ShowError();

ULONG GetNTOSKRNLBase(void)
{


	typedef struct _SYSTEM_MODULE_INFORMATION {
		ULONG Reserved[2];
		PVOID Base;
		ULONG Size;
		ULONG Flags;
		USHORT Index;
		USHORT Unknow;
		USHORT LoadCount;
		USHORT ModuleNameOffset;
		char ImageName[256];        
	} SYSTEM_MODULE_INFORMATION, *PSYSTEM_MODULE_INFORMATION;

	DWORD		(NTAPI *ZwQuerySystemInformation)(ULONG, PVOID, ULONG, PULONG);
	SYSTEM_MODULE_INFORMATION	*ps;
	HANDLE						h;
	DWORD						dwSize;
	DWORD						r,nModules,i;
	UCHAR						*pBuf;
	char						szModuleName[512];
	DWORD						dwBase=0;

	h=GetModuleHandleA("ntdll.dll");

	ZwQuerySystemInformation = (void *)GetProcAddress(h, "ZwQuerySystemInformation");
	if (ZwQuerySystemInformation==0){
		printf("Can not find ZwQuerySystemInformation\n");
		return 0;
	}

	ZwQuerySystemInformation(SystemModuleInformation,&dwSize,0,&dwSize);

	if ((pBuf=malloc(dwSize))==NULL){
		printf("Can not allocate memory (%d bytes)\n",dwSize);
		return 0;
	}

	if ((r=ZwQuerySystemInformation(SystemModuleInformation, pBuf, dwSize, &dwSize))){
		printf("ZwQuerySystemInformation failed r=%d\n",r);
		return 0;
	}
	nModules=*((DWORD *)pBuf);

	ps = (SYSTEM_MODULE_INFORMATION *)(pBuf+4);
	for (i=0;i<nModules;i++){

		memset(szModuleName,0,sizeof(szModuleName));
		strncpy(szModuleName,ps->ImageName,sizeof(szModuleName)-1);
		_strupr(szModuleName);
		if (strstr(szModuleName,"NTKRNLPA.EXE")!=NULL
		||  strstr(szModuleName,"NTOSKRNL.EXE")!=NULL){
			dwBase=(DWORD)(ps->Base);
			free(pBuf);
			return dwBase;
		}

		ps++;
	}
	free(pBuf);
	return 0;
}

int main(int argc,char *argv[])
{
	char			buf[512];
	HANDLE			hFile;
	BOOL			bExit;
	DWORD			dwReturn,dwResult;
	DWORD			dwBase;

	if ((dwBase=GetNTOSKRNLBase())==0){
		printf("Getting base address of Ntoskrnl.exe is failed\n");
		exit(1);
	}

	LoadDriver(DRIVERNAME);
	printf("Load driver completed.\n");
	
	sprintf(buf,"\\\\.\\%s",DEVICENAME);
    if ((hFile = CreateFile(buf, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL))== INVALID_HANDLE_VALUE){
        printf("CreateFile failed. r=%d\n",GetLastError());
		return 3;

	}
	printf("Base of kernel = %x\n",dwBase);
	DeviceIoControl(hFile,IOCTL_ID_CUSTOM_HOOKENABLE,&dwBase,4,&dwBase,4,&dwReturn,NULL);
	
	for (bExit=0;bExit==0;){

		printf("0: End\n");
		printf("1: Add process\n");
		printf("2: Delete process\n");
		printf(">");
		scanf("%s",buf);
		switch(buf[0]){
			case '0':
				bExit=1;
				break;
			case '1':
				printf("Process name : ");
				scanf("%s",buf);
				dwResult=DeviceIoControl(hFile,IOCTL_ID_CUSTOM_ADD,buf,MAXLEN_PROCNAME,buf,MAXLEN_PROCNAME,&dwReturn,NULL);
				if (dwResult==0) ShowError();
				break;
			case '2':
				printf("Process name : ");
				scanf("%s",buf);
				dwResult=DeviceIoControl(hFile,IOCTL_ID_CUSTOM_DELETE,buf,MAXLEN_PROCNAME,buf,MAXLEN_PROCNAME,&dwReturn,NULL);
				if (dwResult==0) ShowError();
				break;
		}
	}


	CloseHandle(hFile);

	UnloadDriver(DRIVERNAME);

	return 0;
}

void ShowError()
{
	DWORD d=GetLastError();
	switch(d){
		case STATUS_ERR_PARAM:
			printf("Error : Parameter error\n");
			break;
		case STATUS_ERR_DUPLICATED:
			printf("Error : Specified process is already in list\n");
			break;
		case STATUS_ERR_LISTFULL:
			printf("Error : List is full\n");
			break;
		case STATUS_ERR_NOTFOUND:
			printf("Error : Specified process is not found\n");
			break;
		default:
			printf("Error : Unknown code %x\n",d);
			break;
	}
}